from typing import Iterable
from abc import ABC
from eric_sse.interfaces import ConnectionRepositoryInterface, ChannelRepositoryInterface, ListenerRepositoryInterface, \
    QueueRepositoryInterface
from eric_sse.entities import AbstractChannel
from eric_sse.connection import Connection
from eric_sse.exception import RepositoryError
from eric_sse.listener import MessageQueueListener
from eric_sse.queues import Queue


class InMemoryStorage:

    def __init__(self, objects: dict[str, any] = None):
        self.objects = objects or {}

    objects: dict[str, any] = {}

    def fetch_by_prefix(self, prefix: str) -> Iterable[any]:
        for k, obj in self.objects.items():
            if k.startswith(prefix):
                yield obj

    def fetch_all(self) -> Iterable[any]:
        for obj in self.objects.values():
            yield obj

    def upsert(self, key: str, value: any):
        self.objects[key] = value

    def fetch_one(self, key: str) -> any:
        try:
            return self.objects[key]
        except KeyError:
            raise RepositoryError(f'Item not found {key}') from None

    def delete(self, key: str):
        del self.objects[key]

class InMemoryRepositoryMixin(ABC):
    def __init__(self):
        self.objects = InMemoryStorage()


class InMemoryQueueRepository(InMemoryRepositoryMixin, QueueRepositoryInterface):

    def load(self, connection_id: str) -> Queue:
        return self.objects.fetch_one(connection_id)

    def persist(self, connection_id: str, queue: Queue):
        self.objects.upsert(connection_id, queue)

    def delete(self, connection_id: str):
        self.objects.delete(connection_id)


class InMemoryListenerRepository(InMemoryRepositoryMixin, ListenerRepositoryInterface):

    def load(self, connection_id: str) -> MessageQueueListener:
        return self.objects.fetch_one(connection_id)

    def persist(self, connection_id: str, listener: MessageQueueListener):
        self.objects.upsert(connection_id, listener)

    def delete(self, connection_id: str):
        self.objects.delete(connection_id)


class InMemoryConnectionRepository(InMemoryRepositoryMixin, ConnectionRepositoryInterface):
    def __init__(self):
        super().__init__()
        self.__queues_repository = InMemoryQueueRepository()
        self.__listeners_repository = InMemoryListenerRepository()

    @property
    def queues_repository(self) -> QueueRepositoryInterface:
        return self.__queues_repository

    @property
    def listeners_repository(self) -> ListenerRepositoryInterface:
        return self.__listeners_repository

    def load_all(self, channel_id: str) -> Iterable[Connection]:
        for connection in self.objects.fetch_all():
            yield connection

    def load_one(self, connection_id: str) -> Connection:
        return self.objects.fetch_one(connection_id)

    def persist(self, connection: Connection):
        self.objects.upsert(connection.id, connection)

    def delete(self, connection_id: str):
        self.objects.delete(connection_id)


class InMemoryChannelRepository(InMemoryRepositoryMixin, ChannelRepositoryInterface):
    def __init__(self, channels: list[AbstractChannel] = None):
        super().__init__()
        for channel in channels or []:
            self.objects.upsert(channel.id, channel)
        self.__connections_repository = InMemoryConnectionRepository()

    @property
    def connections_repository(self) -> ConnectionRepositoryInterface:
        return self.__connections_repository

    def load_all(self) -> Iterable[AbstractChannel]:
        for channel in self.objects.fetch_all():
            yield channel

    def load_one(self, channel_id: str) -> AbstractChannel:
        try:
            return self.objects.fetch_one(channel_id)
        except KeyError:
            raise RepositoryError(f'Item not found {channel_id}') from None

    def persist(self, channel: AbstractChannel):
        self.objects.upsert(channel.id, channel)

    def delete(self, channel_id: str):
        self.objects.delete(channel_id)






from abc import ABC, abstractmethod
from eric_sse.entities import AbstractChannel
from eric_sse.connection import Connection
from eric_sse.interfaces import (ChannelRepositoryInterface, ConnectionRepositoryInterface, ListenerRepositoryInterface,
                                 QueueRepositoryInterface)
from eric_sse.listener import MessageQueueListener
from eric_sse.queues import Queue
from eric_sse.inmemory import (InMemoryChannelRepository, InMemoryConnectionRepository, InMemoryListenerRepository,
                               InMemoryQueueRepository)
from eric_sse.exception import InvalidChannelException

class ApplicationTemplate(ABC):
    def __init__(
            self,
            channel_repository: ChannelRepositoryInterface = InMemoryChannelRepository(),
            connection_repository: ConnectionRepositoryInterface = InMemoryConnectionRepository(),
            listener_repository: ListenerRepositoryInterface = InMemoryListenerRepository(),
            queue_repository: QueueRepositoryInterface = InMemoryQueueRepository(),
    ):
        self._channel_repository = channel_repository
        self._connection_repository = connection_repository
        self._listener_repository = listener_repository
        self._queue_repository = queue_repository
        self._channels: set[str] = set()

    @abstractmethod
    def _create_channel(self) -> AbstractChannel:
        ...

    @abstractmethod
    def _create_listener(self) -> MessageQueueListener:
        ...

    @abstractmethod
    def _create_queue(self) -> Queue:
        ...

    def _create_connection(self) -> Connection:
        return Connection(listener=self._create_listener(), queue=self._create_queue())

    def boot(self) -> dict[str, AbstractChannel]:
        channels: dict[str, AbstractChannel] = {}
        for channel in self._channel_repository.load_all():
            for connection in self._connection_repository.load_all(channel_id=channel.id):
                connection.listener = self._listener_repository.load(connection_id=connection.id)
                connection.queue = self._queue_repository.load(connection_id=connection.id)
                channel.register_connection(connection)
            channels[channel.id] = channel
        return channels

    def create_channel(self) -> AbstractChannel:
        channel = self._create_channel()
        self._channel_repository.persist(channel)
        self._channels.add(channel.id)
        return channel

    def load_channel(self, channel_id: str) -> AbstractChannel:
        channel = self._channel_repository.load_one(channel_id=channel_id)
        self._channels.add(channel.id)
        return channel

    def save_channel(self, channel: AbstractChannel):
        self._channel_repository.persist(channel)
        for connection in channel.get_connections():
            self._save_connection(connection)

    def _save_connection(self, connection: Connection):
        self._connection_repository.persist(connection)
        self._listener_repository.persist(connection.id, connection.listener)
        self._queue_repository.persist(connection.id, connection.queue)

    def delete_channel(self, channel_id: str):
        self._channel_repository.delete(channel_id)
        for connection in self._connection_repository.load_all(channel_id):
            self.unsubscribe_channel(connection.id)
        self._channels.remove(channel_id)

    def subscribe_channel(self, channel: AbstractChannel):
        if channel.id not in self._channels:
            raise InvalidChannelException(f'Unknown channel {channel.id}')

        connection = self._create_connection()
        channel.register_connection(connection)
        self._save_connection(connection)
        return connection

    def unsubscribe_channel(self, channel_id: str, connection_id: str):
        channel = self._channel_repository.load_one(channel_id=channel_id)
        connection = self._connection_repository.load_one(connection_id=connection_id)
        channel.remove_listener(connection.listener.id)
        self._delete_connection(connection_id)

    def _delete_connection(self, connection_id: str):
        self._listener_repository.delete(connection_id=connection_id)
        self._queue_repository.delete(connection_id=connection_id)
        self._connection_repository.delete(connection_id)

    async def listen_to_channel(self, channel: AbstractChannel):
        connection = self.subscribe_channel(channel)
        return channel.message_stream(connection.listener)



import eric_sse
from eric_sse.listener import MessageQueueListener
from eric_sse.queues import Queue

class Connection:
    """
    A connection is just a listener and its related message queue

    :param ~eric_sse.listener.MessageQueueListener listener:
    :param ~eric_sse.queues.Queue queue:
    """
    def __init__(self, listener: MessageQueueListener, queue: Queue, connection_id: str | None = None):
        self.listener = listener
        self.queue = queue
        self.id = connection_id or eric_sse.generate_uuid()
